<?php

if (!defined('ABSPATH')) {
    exit;
}

/**
 * Handles responses from PayPal IPN.
 */
class Easy_PayPal_IPN_Listener {

    /**
     * Constructor.
     */
    public function __construct() {
        $this->liveurl = 'https://ipnpb.paypal.com/cgi-bin/webscr';
        $this->testurl = 'https://ipnpb.sandbox.paypal.com/cgi-bin/webscr';
        $this->wpg_add_log('PayPal IPN Handler initialized.');
    }

    /**
     * Check for PayPal IPN Response.
     */
    public function check_response() {
        $this->wpg_add_log('Checking for PayPal IPN response.');
        if (!empty($_POST) && !empty($_POST['ipn_track_id'])) {
            $this->wpg_add_log('IPN Track ID found: ' . $_POST['ipn_track_id']);
            if (!empty($_POST) && $this->validate_ipn()) {
                $posted = wp_unslash($_POST);
                $this->wpg_add_log('IPN data validated successfully.');
                $this->valid_response($posted);
                exit;
            }
            $this->wpg_add_log('IPN validation failed.');
            wp_die('PayPal IPN Request Failure', 'PayPal IPN', array('response' => 500));
        } else {
            $this->wpg_add_log('No POST data or IPN Track ID found.');
        }
    }

    /**
     * There was a valid response.
     * @param  array $posted Post data after wp_unslash
     */
    public function valid_response($posted) {
        $this->wpg_add_log('Processing valid response.');
        $order = !empty($posted['custom']) ? $this->get_paypal_order($posted['custom']) : false;
        if ($order) {
            $this->wpg_add_log('Order found: #' . $order->get_id());
            $posted['payment_status'] = strtolower($posted['payment_status']);
            if (isset($posted['test_ipn']) && 1 == $posted['test_ipn'] && 'pending' == $posted['payment_status']) {
                $posted['payment_status'] = 'completed';
                $this->wpg_add_log('Test IPN detected. Payment status set to completed.');
            }
            $this->wpg_add_log('Payment status: ' . $posted['payment_status']);
            if (method_exists($this, 'payment_status_' . $posted['payment_status'])) {
                $this->wpg_add_log('Calling payment status handler: payment_status_' . $posted['payment_status']);
                call_user_func(array($this, 'payment_status_' . $posted['payment_status']), $order, $posted);
            } else {
                $this->wpg_add_log('No handler found for payment status: ' . $posted['payment_status']);
            }
        } else {
            $this->wpg_add_log('No order found for the given custom data.');
        }
    }

    /**
     * Check PayPal IPN validity.
     */
    public function validate_ipn() {
        $this->wpg_add_log('Validating IPN response.');
        $validate_ipn = array('cmd' => '_notify-validate');
        $post_log = $_POST;
        $validate_ipn += wp_unslash($_POST);
        $params = array(
            'body' => $validate_ipn,
            'timeout' => 60,
            'httpversion' => '1.1',
            'compress' => false,
            'decompress' => false,
            'user-agent' => 'WooCommerce/' . WC()->version
        );

        $is_sandbox = (isset($_POST['test_ipn'])) ? 'yes' : 'no';
        if ('yes' == $is_sandbox) {
            $paypal_adr = $this->testurl;
            $this->wpg_add_log('Sandbox mode enabled.');
        } else {
            $paypal_adr = $this->liveurl;
            $this->wpg_add_log('Live mode enabled.');
        }

        $response = wp_safe_remote_post($paypal_adr, $params);
        if (!empty($post_log['custom'])) {
            $post_log['custom'] = '*****************';
        }
        $this->wpg_add_log('IPN POST data: ' . print_r($post_log, true));

        if (!is_wp_error($response) && $response['response']['code'] >= 200 && $response['response']['code'] < 300 && strstr($response['body'], 'VERIFIED')) {
            $this->wpg_add_log('Received valid response from PayPal.');
            return true;
        }
        $this->wpg_add_log('Received invalid response from PayPal.');
        if (is_wp_error($response)) {
            $this->wpg_add_log('Error response: ' . $response->get_error_message());
        }
        return false;
    }

    /**
     * Check for a valid transaction type.
     * @param string $txn_type
     */
    public function validate_transaction_type($txn_type) {
        $this->wpg_add_log('Validating transaction type: ' . $txn_type);
        $accepted_types = array('cart', 'instant', 'express_checkout', 'web_accept', 'masspay', 'send_money', 'webaccept');
        if (!in_array(strtolower($txn_type), $accepted_types)) {
            $this->wpg_add_log('Aborting, Invalid type:' . $txn_type);
            exit;
        }
        $this->wpg_add_log('Transaction type validated successfully.');
    }

    /**
     * Check currency from IPN matches the order.
     * @param WC_Order $order
     * @param string $currency
     */
    public function validate_currency($order, $currency) {
        $this->wpg_add_log('Validating currency.');
        $order_currency = $order->get_currency();
        if ($order_currency != $currency) {
            $this->wpg_add_log('Payment error: Currencies do not match (sent "' . $order_currency . '" | returned "' . $currency . '")');
            $order->update_status('on-hold', sprintf(__('Validation error: PayPal currencies do not match (code %s).', 'woo-paypal-gateway'), $currency));
            exit;
        }
        $this->wpg_add_log('Currency validated successfully.');
    }

    /**
     * Check payment amount from IPN matches the order.
     * @param WC_Order $order
     * @param int $amount
     */
    public function validate_amount($order, $amount) {
        $this->wpg_add_log('Validating payment amount.');
        if (number_format($order->get_total(), 2, '.', '') != number_format($amount, 2, '.', '')) {
            $this->wpg_add_log('Payment error: Amounts do not match (gross ' . $amount . ')');
            $order->update_status('on-hold', sprintf(__('Validation error: PayPal amounts do not match (gross %s).', 'woo-paypal-gateway'), $amount));
            exit;
        }
        $this->wpg_add_log('Payment amount validated successfully.');
    }

    /**
     * Handle a completed payment.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_completed($order, $posted) {
        $this->wpg_add_log('Handling completed payment.');
        if ($order->has_status('completed')) {
            $this->wpg_add_log('Aborting, Order #' . $order->get_id() . ' is already complete.');
            exit;
        }
        $this->validate_transaction_type($posted['txn_type']);
        $this->validate_currency($order, $posted['mc_currency']);
        $this->validate_amount($order, $posted['mc_gross']);
        if ('completed' === $posted['payment_status']) {
            $txn_id = '';
            if (!empty($posted['txn_id'])) {
                $txn_id = wc_clean($posted['txn_id']);
            }
            $this->payment_complete($order, $txn_id, sprintf(__('Payment completed successfully via PayPal. Transaction ID: %s', 'woo-paypal-gateway'), $txn_id));
        } else {
            $this->payment_on_hold($order, sprintf(__('Payment pending: %s', 'woo-paypal-gateway'), $posted['pending_reason']));
        }
    }

    /**
     * Handle a pending payment.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_pending($order, $posted) {
        $this->wpg_add_log('Handling pending payment.');
        $this->payment_status_completed($order, $posted);
    }

    /**
     * Handle a failed payment.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_failed($order, $posted) {
        $this->wpg_add_log('Handling failed payment.');
        $order->update_status('failed', sprintf(__('Payment failed via PayPal. Reason: %s', 'woo-paypal-gateway'), wc_clean($posted['payment_status'])));
    }

    /**
     * Handle a denied payment.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_denied($order, $posted) {
        $this->wpg_add_log('Handling denied payment.');
        $this->payment_status_failed($order, $posted);
    }

    /**
     * Handle an expired payment.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_expired($order, $posted) {
        $this->wpg_add_log('Handling expired payment.');
        $this->payment_status_failed($order, $posted);
    }

    /**
     * Handle a voided payment.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_voided($order, $posted) {
        $this->wpg_add_log('Handling voided payment.');
        $this->payment_status_failed($order, $posted);
    }

    /**
     * Handle a refunded order.
     * @param WC_Order $order
     * @param array $posted
     */

    /**
     * Handle a refunded order.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_refunded($order, $posted) {
        $this->wpg_add_log('Processing refund notification.');

        $refund_amount = wc_clean($posted['mc_gross']);
        $order_total = $order->get_total();
        $is_full_refund = ($order_total == ($refund_amount * -1));

        $refund_note = sprintf(
                __('Payment refunded via PayPal. Amount: %s. Reason: %s', 'woo-paypal-gateway'),
                wc_price(abs($refund_amount)),
                isset($posted['reason_code']) ? wc_clean($posted['reason_code']) : __('No reason provided', 'woo-paypal-gateway')
        );

        if ($is_full_refund) {
            $order->update_status('refunded', $refund_note);
        } else {
            $order->add_order_note($refund_note);
        }

        $this->send_ipn_email_notification(
                sprintf(__('Order #%s refunded', 'woo-paypal-gateway'), $order->get_order_number()),
                $refund_note
        );
    }

    /**
     * Handle a reversal.
     * @param WC_Order $order
     * @param array $posted
     */
    public function payment_status_reversed($order, $posted) {
        $this->wpg_add_log('Processing payment reversal.');

        $reason = isset($posted['reason_code']) ? wc_clean($posted['reason_code']) : __('Unknown reason', 'woo-paypal-gateway');
        $note = sprintf(
                __('Payment reversed via PayPal. Reason: %s. Order put on hold.', 'woo-paypal-gateway'),
                $reason
        );

        $order->update_status('on-hold', $note);
        $this->send_ipn_email_notification(
                sprintf(__('Order #%s payment reversed', 'woo-paypal-gateway'), $order->get_order_number()),
                $note
        );
    }

    public function payment_status_canceled_reversal($order, $posted) {
        $this->wpg_add_log('Processing canceled reversal.');

        $note = __('Payment reversal canceled by PayPal. Please verify and update order status if necessary.', 'woo-paypal-gateway');
        $order->add_order_note($note);

        $this->send_ipn_email_notification(
                sprintf(__('Reversal canceled for Order #%s', 'woo-paypal-gateway'), $order->get_order_number()),
                $note
        );
    }

    /**
     * Send a notification to the user handling orders.
     * @param string $subject
     * @param string $message
     */
    public function send_ipn_email_notification($subject, $message) {
        $this->wpg_add_log('Sending IPN email notification.');
        $new_order_settings = get_option('woocommerce_new_order_settings', array());
        $mailer = WC()->mailer();
        $message = $mailer->wrap_message($subject, $message);
        $mailer->send(!empty($new_order_settings['recipient']) ? $new_order_settings['recipient'] : get_option('admin_email'), strip_tags($subject), $message);
    }

    /**
     * Get the order from the PayPal 'Custom' variable.
     * @param  string $raw_custom JSON Data passed back by PayPal
     * @return bool|WC_Order object
     */
    public function get_paypal_order($raw_custom) {
        $this->wpg_add_log('Retrieving order from PayPal custom data.');
        $custom = json_decode($raw_custom);
        if ($custom && is_object($custom)) {
            $order_id = $custom->order_id;
            $order_key = $custom->order_key;
        } else {
            $this->wpg_add_log('Error: Order ID and key were not found in "custom".');
            return false;
        }
        $order = wc_get_order($order_id);
        if (!$order) {
            $order_id = wc_get_order_id_by_order_key($order_key);
            $order = wc_get_order($order_id);
        }
        $order_key_value = $order->get_order_key();
        if (!$order || !hash_equals($order_key_value, $order_key)) {
            $this->wpg_add_log('Error: Order Keys do not match.');
            return false;
        }
        $this->wpg_add_log('Order retrieved successfully: #' . $order->get_id());
        return $order;
    }

    /**
     * Complete order, add transaction ID and note.
     * @param  WC_Order $order
     * @param  string   $txn_id
     * @param  string   $note
     */
    public function payment_complete($order, $txn_id = '', $note = '') {
        $this->wpg_add_log('Completing payment for order #' . $order->get_id());
        $order->add_order_note($note);
        $order->payment_complete($txn_id);
    }

    /**
     * Hold order and add note.
     * @param  WC_Order $order
     * @param  string   $reason
     */
    public function payment_on_hold($order, $reason = '') {
        $this->wpg_add_log('Placing order #' . $order->get_id() . ' on hold.');
        $order->update_status('on-hold', $reason);
        $order->reduce_order_stock();
        WC()->cart->empty_cart();
    }

    /**
     * Add a log entry.
     * @param string $message
     * @param string $level
     */
    public function wpg_add_log($message, $level = 'info') {
        if (empty($this->log)) {
            $this->log = wc_get_logger();
        }
        $this->log->log($level, $message, array('source' => 'woo_paypal_pro'));
    }
}
